package sensor

import (
	"fmt"
	"time"

	"github.com/go-flucky/flucky/pkg/internal/format"
	"github.com/go-flucky/flucky/pkg/types"
	"github.com/go-flucky/go-dht"
	uuid "github.com/satori/go.uuid"
)

// DHT22 is a sensor to measure humidity and temperature.
type DHT22 struct {
	*types.Sensor
}

// GetID returns the sensor id
func (s *DHT22) GetID() string {
	return s.ID
}

// GetTicker returns a new ticker, which tick every when the sensor should be read
func (s *DHT22) GetTicker() *time.Ticker {
	duration, err := time.ParseDuration(s.TickDuration)
	if err != nil {
		duration = time.Minute
	}
	return time.NewTicker(duration)
}

// Read measured values
func (s *DHT22) Read() ([]*types.MeasuredValue, error) {

	err := dht.HostInit()
	if err != nil {
		return nil, fmt.Errorf("HostInit error: %v", err)
	}

	gpio, err := types.GPIOToString(*s.GPIONumber)
	if err != nil {
		return nil, err
	}

	dht, err := dht.NewDHT(gpio, dht.Celsius, "")
	if err != nil {
		return nil, fmt.Errorf("NewDHT error: %v", err)
	}

	humidityValue, temperatureValue, err := dht.Read()
	if err != nil {
		return nil, fmt.Errorf("Read error: %v", err)
	}

	measuredValues := []*types.MeasuredValue{
		&types.MeasuredValue{
			ID:        uuid.NewV4().String(),
			Value:     float64(humidityValue),
			ValueType: types.MeasuredValueTypeHumidity,
			FromDate:  format.FormatedTime(),
			TillDate:  format.FormatedTime(),
			SensorID:  s.ID,
		},
		&types.MeasuredValue{
			ID:        uuid.NewV4().String(),
			Value:     float64(temperatureValue),
			ValueType: types.MeasuredValueTypeTemperature,
			FromDate:  format.FormatedTime(),
			TillDate:  format.FormatedTime(),
			SensorID:  s.ID,
		},
	}

	return measuredValues, nil
}
